/* 
 * Primitives de dessin pour libNDS
 * Auteur : Fares BELHADJ
 * Email  : amsi@ai.univ-paris8.fr
 * Date   : 02/10/2007 modifie le 09/10/2007
 */

#include <math.h>
#include <stdlib.h>
#include <assert.h>

#include "primitives.h"

/*
 * Couleur en cours
 */
static uint16 cur_color15 = 0x0;

/*
 * Modifier la couleur 15 bits en cours
 */
extern uint16 setColor15(uint16 c) {
  return cur_color15 = c;
}

/*
 * Renvoie la couleur 15 bits en cours
 */
extern uint16 getColor15(uint16 c) {
  return cur_color15;
}

/*
 * dessine un pixel avec cur_color dans VRAM_A
 */
extern void vramaPutPixel(int x, int y) {
  if(inScreen(x, y))
    VRAM_A[y * SCREEN_WIDTH + x] = cur_color15;
}

/*
 * line dessine un segment (x0, y0) -> (x1, y1)
 * en utilisant l'algorithme de Bresenham'65.
 */
extern void line(void (* ppfunc)(int x, int y), int x0, int y0, int x1, int y1) {
  int u = x1 - x0, v = y1 - y0;
  int pasX = (u < 0) ? -1 : 1, pasY = (v < 0) ? -1 : 1;
  int x, y, del, incH, incO;
  if(abs(u) < abs(v)) { /* deuxieme octant */
    del = (incH = ((pasX * u) << 1)) - pasY * v;
    incO = incH - ((pasY * v) << 1);
    for(y = y0, x = x0; y != y1; y += pasY) {
      ppfunc(x, y);
      if(del < 0) del += incH;
      else {
        del += incO;
        x += pasX;
      }
    }
  } else {  /* premier octant */
    del = (incH = ((pasY * v) << 1)) - pasX * u;
    incO = incH - ((pasX * u) << 1);
    for(x = x0, y = y0; x != x1; x += pasX) {
      ppfunc(x, y);
      if(del < 0) del += incH;
      else {
        del += incO;
        y += pasY;
      }
    }
  }
}

/*
 * line dessine un segment horizontal avec degrade de p a q.
 */
extern void shadedHLine(void (* ppfunc)(int x, int y), vertex2i3u_t * p, vertex2i3u_t * q, int y) {
  vertex2i3u_t * t;
  int x, x0, x1, c, d;
  if(!(q->x - p->x)) {
    if(inScreen(p->x, y)) {
      setColor15(RGB15(p->r, p->g, p->b));
      ppfunc(p->x, y);
    }
    return;
  }
  if(y < 0 || y >= SCREEN_HEIGHT) return;
  if(q->x < p->x) { t = p; p = q; q = t; }
  x0 = MAX(p->x, 0); x1 = MIN(q->x, SCREEN_WIDTH - 1);
  d = x1 - x0;
  for(c = 0, x = x0; x <= x1; x++, c++) {
    setColor15(	RGB15( ((d - c) * p->r + c * q->r) / d,
		       ((d - c) * p->g + c * q->g) / d,
		       ((d - c) * p->b + c * q->b) / d
		       ) 
		);
    ppfunc(x, y);
  }
}

/*
 * shadedScanline dessine un polygone  n cotes et fait le remplissage
 * avec un degrade entre les couleurs donnees en chaque sommet.
 */
extern void shadedScanline(void (* ppfunc)(int x, int y), vertex2i3u_t * p, int n) {
  int i, j, oi, oj, k, imin = 0, imax = 0, h, sg, sd, dg, dd;
  vertex2i3u_t * xg, * xd;
  for(i = 1; i < n; i++) {
    if(p[imin].y > p[i].y) imin = i;
    if(p[imax].y < p[i].y) imax = i;
  }
  xg = malloc((h = p[imax].y - p[imin].y + 1) * sizeof * xg);
  xd = malloc(h-- * sizeof * xd);
  assert(xg && xd);
  for(oi = oj = i = j = 0, dg = ((sg = sd = imin) + 1) % n, dd = (sd + n - 1) % n; i < h || j < h;) {
    if(i < h) i += v2i3uAbscisses(&p[sg], &p[dg], &xg[i]);
    if(j < h) j += v2i3uAbscisses(&p[sd], &p[dd], &xd[j]);
    for(k = MIN(oi, oj); k <= MIN(i, j); k++)
      shadedHLine(ppfunc, &xg[k], &xd[k], k + p[imin].y);
    oi = i; oj = j; sg = dg; sd = dd;
    dg = (dg + 1) % n; dd = (dd + n - 1) % n; 
  }
  free(xg); free(xd);
}

/*
 * calcule les abscisses entre p et q et les stoque a partir de a.
 */
extern int  v2i3uAbscisses(vertex2i3u_t * p, vertex2i3u_t * q, vertex2i3u_t * a) {
  float f1, f2;
  int u = q->x - p->x, v = q->y - p->y; /* ici v est tjs >= 0 */
  int x, y, pasx = u > 0 ? 1 : -1, delta, inch, inco;
  if (abs(u) < v) {
    delta = (inch  = (pasx * u) << 1) - v;
    inco  = (pasx * u - v) << 1;
    for( y = 0, x = p->x; y <= v; y++) {
      a[y].x = x;
      f1 = (v - y + fabs(q->x - x)) / (float)(abs(u) + v);
      f2 = (y + fabs(x - p->x))     / (float)(abs(u) + v);
      a[y].r =  f1 * p->r + f2 * q->r;
      a[y].g =  f1 * p->g + f2 * q->g;
      a[y].b =  f1 * p->b + f2 * q->b;
      if(delta < 0) delta += inch;
      else { delta += inco; x += pasx; }
    }
  } else {
    delta = (inch  = (v << 1)) - pasx * u;
    inco  = (v - pasx * u) << 1;
    if(!v || !u) { a[0] = *q; return v; }
    for( y = 0, x = p->x; x != q->x + pasx; x += pasx) {
      a[y].x = x;
      f1 = (v - y + fabs(q->x - x)) / (float)(abs(u) + v);
      f2 = (y + fabs(x - p->x))     / (float)(abs(u) + v);
      a[y].r =  f1 * p->r + f2 * q->r;
      a[y].g =  f1 * p->g + f2 * q->g;
      a[y].b =  f1 * p->b + f2 * q->b;
      if(delta < 0) delta += inch;
      else { delta += inco; y++; }
    }
  }
  return v;
}
